﻿using System;
using System.Linq;
using HeuristicLab.Core;
using HeuristicLab.Services.Hive.Common;
using HeuristicLab.Services.Hive.Common.DataTransfer;

namespace HeuristicLab.Services.Hive {
  /// <summary>
  /// This class holds the state of all recent heartbeats and decides to reschedule jobs and set slaves offline
  /// </summary>
  public class LifecycleManager : ILifecycleManager {
    private DataAccess.IHiveDao dao {
      get { return ServiceLocator.Instance.HiveDao; }
    }
    private HeuristicLab.Services.Hive.DataAccess.TransactionManager trans {
      get { return ServiceLocator.Instance.TransactionManager; }
    }
    private IAuthorizationManager auth {
      get { return ServiceLocator.Instance.AuthorizationManager; }
    }

    private static object locker = new object();

    // Windows-Forms timer is single threaded, so callbacks will be synchron
    private System.Windows.Forms.Timer timer;

    public ExecutionState ExecutionState {
      get { return timer.Enabled ? ExecutionState.Started : Core.ExecutionState.Stopped; }
    }

    public LifecycleManager() {
      this.timer = new System.Windows.Forms.Timer();
      this.timer.Tick += new EventHandler(timer_Tick);
    }

    public void Start() {
      if (ExecutionState == Core.ExecutionState.Stopped) {
        // kick off the event immediately
        timer_Tick(this, EventArgs.Empty);

        // start the timer
        this.timer.Interval = (int)new TimeSpan(0, 0, 30).TotalMilliseconds;
        this.timer.Start();
      }
    }

    public void Stop() {
      if (ExecutionState == Core.ExecutionState.Started) {
        timer.Stop();
      }
    }

    /// <summary>
    /// This method is supposed to check if slaves are online
    /// if not -> set them offline and check if they where calculating a job
    /// </summary>
    private void timer_Tick(object sender, EventArgs e) {
      lock (locker) {
        using (trans.OpenTransaction()) {
          SetTimeoutSlavesOffline();
          FinishParentJobs();
        }
      }
    }

    /// <summary>
    /// Searches for slaves which are timed out, puts them and their jobs offline
    /// </summary>
    private void SetTimeoutSlavesOffline() {
      var slaves = dao.GetSlaves(x => x.SlaveState != SlaveState.Offline);
      foreach (Slave slave in slaves) {
        if (!slave.LastHeartbeat.HasValue || (DateTime.Now - slave.LastHeartbeat.Value).TotalSeconds > ApplicationConstants.HeartbeatTimeout) {
          slave.SlaveState = SlaveState.Offline;
          SetJobsWaiting(slave.Id);
          dao.UpdateSlave(slave);
        }
      }
    }

    /// <summary>
    /// Looks for parent jobs which have FinishWhenChildJobsFinished and set their state to finished
    /// </summary>
    private void FinishParentJobs() {
      var parentJobsToFinish = dao.GetParentJobs(dao.GetResources(x => true).Select(x => x.Id), 0, true);
      foreach (var job in parentJobsToFinish) {
        job.SetState(JobState.Finished);
        dao.UpdateJob(job);
      }
    }

    private void SetJobsWaiting(Guid slaveId) {
      var jobs = dao.GetJobs(x => x.State == JobState.Calculating).Where(x => x.StateLog.Last().SlaveId == slaveId);
      foreach (var j in jobs) {
        j.StateLog.Add(new StateLog() { 
          State = JobState.Waiting,
          JobId = j.Id,
          DateTime = DateTime.Now,
          Exception = "Slave timed out"
        });
        dao.UpdateJob(j);
      }
    }
  }
}
